from conftest import URL, JOB_ID_SUCCESS, JOB_ID_ERROR, TEST_TRANS_SQL, TEST_TRANS_PY
from pykeboola.jobs import JobsClient
from datetime import datetime

def test_check_job_id(token):
    jobs = JobsClient(URL, token)
    assert jobs.check_job_status(JOB_ID_ERROR) == 'error'
    assert jobs.check_job_status(JOB_ID_SUCCESS) == 'success'

def test_queue_job(token):
    jobs = JobsClient(URL, token)
    values = [{
        "name": "test_datetime",
        "value": f"'{datetime.now().strftime('%Y-%m-%d %H:%M:%S')}'"
    },
    {
        "name": "test_version",
        "value": f"1"
    }]
    job_id_sql = jobs.queue_job('keboola.snowflake-transformation',TEST_TRANS_SQL, values)
    job_id_py = jobs.queue_job('keboola.python-transformation-v2',TEST_TRANS_PY, values)
    assert int(job_id_sql) > 0
    assert int(job_id_py) > 0
    


import requests
from dataclasses import dataclass, field, fields
from typing import List, Dict

@dataclass
class Table:
    id: str = field(compare=False)
    name: str
    schema: str
    schema_id: str = field(compare=False)
    description: str = field(default=None)
    row_cnt: str = field(default=None, compare=False)
    columns: List = field(default_factory=list)
    native_types: bool = field(default=None, compare=False)

    @property
    def primary_keys(self) -> List: 
        return [column for column in self.columns if column.primary]
    
    @classmethod
    def from_keboola(cls, keboola_json: Dict):
        return cls(
            id = keboola_json['id'],
            name = keboola_json['displayName'],
            schema = keboola_json['bucket']['displayName'],
            schema_id = keboola_json['bucket']['id'],
            description = next(iter([meta['value'] for meta in keboola_json.get('metadata') or [] if meta['key'] == 'KBC.description']), None),
            row_cnt = keboola_json.get('rowsCount'),
            columns = Column.from_keboola(keboola_json.get('columnMetadata'), keboola_json.get('primaryKey'), keboola_json.get('columns')),
            native_types = keboola_json.get('isTyped')
        )
    
    @classmethod
    def new_table(cls, name, schema_id, columns):
        return cls(
            id = None,
            name = name,
            schema = None,
            schema_id = schema_id,
            columns = columns
        )
    
    def __sub__(self, other):
        differences = {}
        for field in fields(self):
            if field.compare and getattr(self, field.name) != getattr(other, field.name):
                differences[field.name] = (getattr(self, field.name), getattr(other, field.name))
        
        if 'columns' in differences:
            col_dif = {}
            self_col_dict = {column.name: column for column in self.columns}
            other_col_dict = {column.name: column for column in other.columns}
            for col_name in self_col_dict:
                if col_name in other_col_dict:
                    dif = self_col_dict[col_name] - other_col_dict[col_name]
                    if dif:
                        col_dif[col_name] = dif
                else:
                    col_dif[col_name] = 'Left Only'
            for col_name in other_col_dict:
                if not col_name in self_col_dict:
                    col_dif[col_name] = 'Right Only'
            if col_dif:
                differences['columns'] = col_dif
            else:
                differences.pop('columns')
        
        return differences

@dataclass
class Column:
    name: str
    type: str
    description: str = field(default=None)
    primary: bool = field(default=False)
    length: str = field(default=None)

    @classmethod
    def from_keboola(cls, column_metadata, primary_columns, columns) -> List:
        cols = []
        if column_metadata:
            for name, metadata in column_metadata.items():
                cols.append(cls(
                    name = name,
                    type = next(iter([meta['value'] for meta in metadata if meta['key'] == 'KBC.datatype.basetype' and meta['provider'] == 'storage']), None),
                    length = next(iter([meta['value'] for meta in metadata if meta['key'] == 'KBC.datatype.length' and meta['provider'] == 'storage']), None),
                    description = next(iter([meta['value'] for meta in metadata if meta['key'] == 'KBC.description']), None),
                    primary = name in primary_columns
                ))
        elif columns:
            for name in columns:
                cols.append(cls(
                    name = name,
                    type = None,
                    length = None,
                    description = None,
                    primary = name in primary_columns,
                ))
        return cols
    
    def __sub__(self, other):
        differences = {}
        for field in fields(self):
            if field.compare and getattr(self, field.name) != getattr(other, field.name):
                differences[field.name] = (getattr(self, field.name), getattr(other, field.name))
        return differences
    

class TablesClient:
    url: str
    token: str
    headers: dict
    def __init__(self, base_url: str, token: str):
        self.token = token
        self.url = f"{base_url.rstrip('/')}/v2/storage"
        self.headers = {
            'Content-Type': 'application/json',
            'X-StorageApi-Token': self.token
        }
    
    def get_tables(self, bucket: str = None) -> List[Table]:
        if bucket:
            url = f"{self.url}/buckets/{bucket}/tables?include=columns,buckets,metadata,columnMetadata"
        else:
            url = f"{self.url}/tables?include=columns,buckets,metadata,columnMetadata"
        response = requests.get(url, headers=self.headers)
        if response.status_code == 200:
            return [Table.from_keboola(table_json) for table_json in response.json()]
        else:
            err = response.text
            raise requests.HTTPError(f'Failed to retrieve tables from Keboola: {err}')
        
    def create_table(self, table: Table):
        url = f'{self.url}/buckets/{table.schema_id}/tables-definition'
        values = {
            "name": table.name,
            "primaryKeysNames": [column.name for column in table.primary_keys],
            "columns": [{"name": column.name,
                         "definition":
                         {
                             "type": column.type,
                             "length": column.length,
                         }} for column in table.columns]
        }
        response = requests.post(url, headers=self.headers, json=values)
        if response.status_code == 202:
            return response.json()['id']
        else:
            err = response.text
            raise requests.HTTPError(f'Failed to create table in Keboola: {err}\nBody: {values}')
            

    def delete_table(self, table_id):
        url = f'{self.url}/tables/{table_id}/?force=true'
        response = requests.delete(url, headers=self.headers)
        if response.status_code == 204:
            return True
        else:
            err = response.text
            raise requests.HTTPError(f'Failed to delete table {table_id} in Keboola: {err}')
        
    def update_table_metadata(self, table: Table):
        url = f'{self.url}/tables/{table.id}/metadata'
        values = {
            'provider': 'user',
        }
        if table.description:
            values['metadata'] = [
                {
                    'key': 'KBC.description',
                    'value': table.description
                }
            ]

        if table.columns:
            values['columnsMetadata'] = {
                column.name: [{
                    'key': 'KBC.description', 
                    'value': column.description
                    }]
                for column in table.columns}
        
        response = requests.post(url, headers=self.headers, json=values)
        if response.status_code == 201:
            return response.json()
        else:
            err = response.text
            raise requests.HTTPError(f'Failed to add metadata to table in Keboola: {err}\nBody: {values}')


from pykeboola.jobs import JobsClient
from pykeboola.tables import TablesClient

class Client:
    """
    Object which hold the base url for Keboola and redirects user to individual objects
    to interact with via API calls.
    """
    base_url: str
    token: str
    jobs: JobsClient
    tables: TablesClient
    def __init__(self, base_url: str, token: str):
        self.base_url = base_url.rstrip('/')
        self.token = token
        self.jobs = JobsClient(base_url, token)
        self.tables = TablesClient(base_url, token)



import requests
from typing import Dict

class JobsClient:
    """
    Class to interact with jobs in Keboola queue.
    """
    queue_url: str
    token: str

    def __init__(self, base_url: str, token: str):
        self.queue_url = base_url.replace('connection', 'queue').rstrip('/')
        self.storage_url = f'{base_url.rstrip("/")}/v2/storage'
        self.token = token

    def queue_job(self, component_id: str, configuration_id: int, variable_values: dict = None, branch_id: int = None):
        """
        Queues a job in Keboola. Returns job_id and raises error on response codes >= 400.
        """
        body = {
        'component': component_id,
        'config': configuration_id,
        'mode': 'run',
        }
        if variable_values:
            body['variableValuesData'] = {'values': variable_values}
        if branch_id:
            body['branchId'] = branch_id
        headers = {
            'X-StorageApi-Token': self.token
        }
        response = requests.post(f'{self.queue_url}/jobs', headers=headers, json=body)
        if response.status_code >= 400:
            err = response.text
            raise requests.HTTPError(f"Failed to queue a job in Keboola. API Response: {err}")
        
        return response.json()['id']
    
    def check_job_status(self, job_id) -> str:
        """
        Checks the status of the job. Returns current status.
        """
        return self.get_job(job_id)['status']
    
    def get_job(self, job_id) -> Dict:
        """
        Gets all info about a queue job.
        """
        url = f'{self.queue_url}/jobs/{job_id}'
        headers = {
            'Content-Type': 'application/json',
            'X-StorageApi-Token': self.token
        }
        response = requests.get(url, headers=headers)
        if response.status_code >= 400:
            err = response.text
            raise requests.HTTPError(f"Failed to find a job in Keboola. API Response: {err}")
        return response.json()
    
    def check_api_job_status(self, job_id) -> str:
        """
        Checks the status of the job. Returns current status.
        """
        return self.get_api_job(job_id)['status']

    def get_api_job(self, job_id):
        """
        Gets all info about an API job.
        """
        url = f'{self.storage_url}/jobs/{job_id}'
        headers = {
            'Content-Type': 'application/json',
            'X-StorageApi-Token': self.token
        }
        response = requests.get(url, headers=headers)
        if response.status_code >= 400:
            err = response.text
            raise requests.HTTPError(f"Failed to find a job in Keboola. API Response: {err}")
        return response.json()



from pykeboola.tables import TablesClient, Table, Column
from pykeboola import Client
from conftest import URL
import json
import time

def test_get_tables(token):
    tables = TablesClient(URL, token)
    assert type(tables.get_tables()) == list
    assert type(tables.get_tables()[0]) == Table

def test_table_from_keboola():
    with open('tests/test_tables.json', 'r') as reader:
        json_list = json.loads(reader.read())
    
    for table_json in json_list:
        table = Table.from_keboola(table_json)
        assert table.id == table_json['id']
        assert table.name == table_json['displayName']
        assert table.schema_id == table_json['bucket']['id']
        assert table.schema == table_json['bucket']['displayName']
        assert table.description == next(iter([meta['value'] for meta in table_json['metadata'] or [] if meta['key'] == 'KBC.description']), None)
        assert table.row_cnt == table_json['rowsCount']
        assert table.native_types == table_json['isTyped']
        assert table.columns == Column.from_keboola(table_json['columnMetadata'], table_json['primaryKey'], table_json['columns'])

def test_table_from_keboola_no_meta():
    with open('tests/test_table_no_meta.json', 'r') as reader:
        json_dict = json.loads(reader.read())
    table = Table.from_keboola(json_dict)
    assert table.name == 'BULK_EXEC_TEST'
    assert table.schema == 'BULK_EXEC_TEST'
    assert table.description == None
    assert table.row_cnt == 1
    assert table.columns == []
    assert table.primary_keys == []

def test_table_create_delete_api_job(token):
    client = Client(URL, token)
    col1 = Column('TEST_COL1', 'VARCHAR', primary=True)
    col2 = Column('TEST_COL2', 'DATETIME')
    table = Table(
        id = None,
        name = 'PYTEST_TEST',
        schema_id = 'out.c-NATIVE_TABLE_TEST',
        schema = None,
        columns=[col1, col2],
    )
    job_id = client.tables.create_table(table)
    status = client.jobs.check_api_job_status(job_id)
    timeout = 10
    timer = 0
    while status != 'success':
        if status == 'error':
            raise ValueError('Table creation returned an error')
        
        timer += 1
        if timer > timeout:
            raise TimeoutError('Waiting for table creation is timing out.')
        time.sleep(1)
        status = client.jobs.check_api_job_status(job_id)
    delete_status = client.tables.delete_table(f'{table.schema_id}.{table.name}')
    assert delete_status == True
    
